"""
grib_processor.main
~~~~~~~~~~~~~~~~~~~

Unified entrypoint to the grib processor.
"""

from __future__ import annotations


import argparse
from itertools import chain
import logging
import logging.handlers
import os
import signal
import sys
from typing import Callable, Iterable
import warnings

from dotenv import load_dotenv
import ssec_amqp.api as mq

from grib_processor.grib import GribPayload, itergrib, load_xcd_models
from grib_processor.utils import (
    dump_message,
    grib_file_watch,
    publish_message,
    realtime,
    DEFAULT_WATCH_DEBOUNCE,
    REALTIME_WATCH_DEBOUNCE,
    signalcontext,
)


# Logging stuff
LOG = logging.getLogger("grib_ingest")
LOG_LEVELS = [
    logging.CRITICAL,
    logging.ERROR,
    logging.WARNING,
    logging.INFO,
    logging.DEBUG,
]
LOG_NAME = "grib_processor.log"

# Where we're publishing data to
DEFAULT_AMQP_SERVERS = ["mq1.ssec.wisc.edu", "mq2.ssec.wisc.edu", "mq3.ssec.wisc.edu"]

# Where on the servers we're putting the data.
DEFAULT_AMQP_EXCHANGE = "model"


def initialize_logging(verbosity: int, rotating_dir: str | None = None) -> None:
    """Set up logging for the package. Don't use basicConfig so we don't
    override other's logs.

    Optionally, can specify an existing directory to put rotating log files into.
    """
    # Set up logging
    log_formatter = logging.Formatter(
        fmt="[%(asctime)s][%(levelname)s]-%(message)s", datefmt="%Y-%m-%dT%H:%M:%S"
    )

    if rotating_dir is not None:
        file_handler = logging.handlers.TimedRotatingFileHandler(
            os.path.join(rotating_dir, LOG_NAME), when="D", utc=True
        )
        file_handler.setFormatter(log_formatter)
        file_handler.setLevel(logging.DEBUG)
        LOG.addHandler(file_handler)

    out_handler = logging.StreamHandler()
    out_handler.setFormatter(log_formatter)
    out_handler.setLevel(LOG_LEVELS[min(verbosity, len(LOG_LEVELS) - 1)])
    LOG.addHandler(out_handler)

    LOG.setLevel(logging.DEBUG)


def setup() -> tuple[Iterable[GribPayload], Callable[[GribPayload], None]]:
    """Initialize the script so it's ready to run."""
    # Set up cmdline args
    parser = argparse.ArgumentParser(
        description="Set paramters for ingesting and publishing of grib data."
    )
    parser.add_argument(
        "grib_src",
        nargs="?",
        default=None,
        metavar="grib_source",
        help="Where to get grib files from. Either stdin (default), a directory to watch, or a grib file itself.",
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        type=int,
        default=1,
        help="Specify verbosity to stderr from 0 (CRITICAL) to 5 (DEBUG).",
    )
    parser.add_argument(
        "--log-dir",
        default=None,
        help="Set up a rotating file logger in this directory.",
    )

    parser.add_argument(
        "-R",
        "--realtime",
        action="store_true",
        help="Turns on parameters for more efficient realtime processing like caching and debounce. Also adds more robust error handling.",
    )
    parser.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="When using a directory as the source, process all files in the directory recursively instead of setting up a watch.",
    )

    parser.add_argument(
        "-o",
        "--output",
        choices=["amqp", "json"],
        default="json",
        help="Where to output processed grib messages to. Default is %(default)s.",
    )

    mq_group = parser.add_argument_group(
        "RabbitMQ arguments",
        "Options for how to output grib messages when using '-o=amqp'.",
    )
    mq_group.add_argument(
        "--servers",
        nargs="+",
        default=DEFAULT_AMQP_SERVERS,
        help="Servers to publish AMQP messages to. Defaults to %(default)s",
    )
    mq_group.add_argument(
        "-X",
        "--exchange",
        default=DEFAULT_AMQP_EXCHANGE,
        help='RabbitMQ exchange to publish the messages to. Defaults to "%(default)s"',
    )
    mq_group.add_argument(
        "-u",
        "--user",
        default=None,
        help="RabbitMQ user to login with. Can also specify using a .env file, see README#configuration.",
    )
    mq_group.add_argument(
        "-p",
        "--password",
        default=None,
        help="RabbitMQ password to login with. Can also specify using a .env file, see README#configuration.",
    )

    args = parser.parse_args()

    if args.verbosity < 0:
        parser.error("--verbosity cannot be negative!")
    initialize_logging(args.verbosity, args.log_dir)
    load_xcd_models()

    # Get an iterator over grib files to process
    file_iter: Iterable[str]
    if args.grib_src is None:
        LOG.info("Sourcing grib files from stdin, realtime=%s", args.realtime)
        file_iter = map(str.strip, iter(sys.stdin.readline, ""))
    elif os.path.isdir(args.grib_src):
        LOG.info(
            "Sourcing grib files from directory watch @ %s realtime=%s",
            args.grib_src,
            args.realtime,
        )
        if not args.realtime:
            warnings.warn("Processing files from a directory without using --realtime!")
            debounce = DEFAULT_WATCH_DEBOUNCE
        else:
            debounce = REALTIME_WATCH_DEBOUNCE

        file_iter = grib_file_watch(
            args.grib_src, debounce=debounce, recursive=args.recursive
        )
    elif os.path.isfile(args.grib_src):
        LOG.info("Grib source is a file, parsing directly, %s", args.grib_src)
        file_iter = (str(args.grib_src),)
    else:
        parser.error("{0} is not a valid source!".format(args.grib_src))

    # Get an iterator of grib messages to emit
    source: Iterable[GribPayload]
    if args.realtime:
        source = realtime(file_iter)
    else:
        source = chain.from_iterable(map(itergrib, file_iter))

    # Possible to source connection parameters from the environment
    if not load_dotenv():
        LOG.warning("Couldn't find .env file")

    if args.output == "amqp":
        LOG.info("Emitting messages to amqp")
        # will automatically disconnect on exit
        mq.connect(
            *args.servers,
            user=args.user or os.getenv("RMQ_USER", None),
            password=args.password or os.getenv("RMQ_PASS", None),
            exchange=args.exchange,
        )
        LOG.info("Connected to %s", ", ".join(args.servers))
        emit = publish_message
    else:
        LOG.info("Emitting messages to stdout")
        emit = dump_message

    return source, emit


def __raise_interrupt(_sig, _frame):  # noqa: ARG001
    """Signal handler that raises KeybardInterrupt."""
    raise KeyboardInterrupt


def main() -> None:
    """Main method, setup and start processing."""
    try:
        source, emit = setup()
    except KeyboardInterrupt:
        return

    with signalcontext(signal.SIGTERM, __raise_interrupt):
        try:
            for grib_msg in source:
                LOG.debug("Got %s from source", grib_msg)
                emit(grib_msg)
        except KeyboardInterrupt:
            LOG.critical("Got interrupt, goodbye!")
    return
