"""
grib_processor.sources
~~~~~~~~~~~~~~~~~~~~~~

The different ways to source grib files for processing.
"""

from __future__ import annotations

import atexit
from collections import defaultdict
from contextlib import contextmanager
import json
import logging
import os
import signal
import time
from typing import Callable, DefaultDict, Generator, Iterable, Tuple

from watchfiles import watch, Change
from typing_extensions import TypeAlias
import ssec_amqp.api as mq

from grib_processor.grib import GribPayload, itergrib


# Our logger
LOG = logging.getLogger("grib_ingest")

# How long grib files stay in the message cache.
CACHE_TTL = 43200  # (12 hours)

# How often to clean the cache of previous messages.
CACHE_CLEAN_INTERVAL = 5400  # (1.5 hours)

# What file extensions to parse as grib files
GRIB_ENDINGS = [".grb", ".grib", ".grb2", ".grib2"]

# Debounce is how long to group grib file changes over before processing them

# watchfiles default debounce
DEFAULT_WATCH_DEBOUNCE = 1600

# debounce when doing realtime processing
REALTIME_WATCH_DEBOUNCE = 4500

# Format of the route key to messages publish with
GRIB_ROUTE_KEY_FMT = "{xcd_model}.{xcd_model_id}.realtime.reserved.reserved.2"

# Cache path
REALTIME_CACHE_PATH = os.path.join(
    os.path.expanduser("~/.cache/"), "grib_processor_cache.json"
)

# Types used for realtime caching
_CEntry: TypeAlias = Tuple[float, int]
_Cache: TypeAlias = DefaultDict[str, _CEntry]


@contextmanager
def signalcontext(signum: int, handler: Callable):
    """Context manager that changes a signal handler on entry and resets it
    on exit.

    Args:
        signum (int): Signal to change.
        handler (tp.Callable): New signal handler to use.
    """
    try:
        orig_h = signal.signal(signum, handler)
        yield
    finally:
        signal.signal(signum, orig_h)


def _default_cache_entry() -> _CEntry:
    """default cache entry (current time, first grib message)."""
    return time.time(), 0


def _initialize_cache() -> _Cache:
    """Sets up the realtime cache to use. Potentially loading previously saved
    caches.

    Returns:
        _Cache: Cache of file names to last grib message processed.
    """
    LOG.info("Loading cache from disk")
    try:
        with open(REALTIME_CACHE_PATH, "r", encoding="utf-8") as cache_file:
            cache_data = json.load(cache_file)
            LOG.info("realtime - cache loaded from disk at %s", REALTIME_CACHE_PATH)
    except (OSError, json.JSONDecodeError):
        LOG.exception("Error loading previous cache data from %s", REALTIME_CACHE_PATH)
        cache_data = {}

    return defaultdict(_default_cache_entry, cache_data)


def _save_cache(cache: _Cache) -> None:
    """Saves a cache to disk at the path REALTIME_CACHE_PATH.

    Args:
        cache (_Cache): The cache to save to disk.
    """
    LOG.info("realtime - saving cache to disk")
    if not cache:
        LOG.debug("realtime - cache is empty")
        return
    try:
        with open(REALTIME_CACHE_PATH, "w", encoding="utf-8") as cache_file:
            json.dump(cache, cache_file)
            LOG.info("realtime - cache saved to disk at %s", REALTIME_CACHE_PATH)
    except (OSError, ValueError):
        LOG.exception("Couldn't save cache to %s", REALTIME_CACHE_PATH)


def _clean_caches(realtime_cache: _Cache) -> None:
    """Removes expired items from the cache. Also, cleans caches
    used by 3rd party libraries (grib2io) that could cause memory leaks
    in long-running processing.

    Args:
        cache (_Cache): The realtime cache to clean.
    """

    to_remove = []
    LOG.debug("realtime - Cleaning the cache, size=%d!", len(realtime_cache))
    for path, (last_update_time, last_msg_num) in realtime_cache.items():
        time_in_cache = time.time() - last_update_time
        if time_in_cache > CACHE_TTL:
            LOG.info(
                "realtime - Evicted %s from cache! (last message: %d, time since last update: %f).",
                path,
                last_msg_num - 1,
                time_in_cache,
            )
            to_remove.append(path)
    if not to_remove:
        LOG.debug("realtime - Nothing to remove from cache...")
    else:
        for rem_path in to_remove:
            realtime_cache.pop(rem_path, None)
        # LOG.debug('Cleaning grib2io caches')
        # grib2io._grib2io._msg_class_store.clear()
        # grib2io._grib2io._latlon_datastore.clear()


def realtime(file_iter: Iterable[str]) -> Generator[GribPayload, None, None]:
    """Add message caching and exception handling to a grib file iterator. This
    is useful if the iterator returns files that aren't fully written to or
    duplicates.

    Args:
        file_iter (Iterator[str]): An iterator returning paths to grib files.

    Yields:
        Generator[GribMessage, None, None]: A GribMessage generator.
    """
    cache = _initialize_cache()
    _clean_caches(cache)
    last_clean_time = time.time()
    atexit.register(_save_cache, cache)

    for file in file_iter:
        LOG.info("realtime - got file %s", file)
        next_msg = cache.get(file, (None, None))[1]
        if next_msg is None:
            LOG.debug("realtime - cache miss %s", file)
            cache[file] = _default_cache_entry()
            next_msg = 0
        else:
            LOG.debug("realtime - cache hit %s @ %d", file, next_msg)
        msg_num = None
        try:
            for msg_num, msg in enumerate(
                itergrib(file, start_message=next_msg), next_msg
            ):
                LOG.debug("realtime - got msg %s:%d", file, msg_num)
                yield msg
        except Exception:
            LOG.exception(
                "realtime - Error processing grib file %s @ msg %d", file, msg_num or 0
            )

        if msg_num is not None:  # msg_num is None when exception or end of grib file
            cache[file] = time.time(), msg_num + 1

        if time.time() - last_clean_time > CACHE_CLEAN_INTERVAL:
            LOG.info("realtime - time to clean the cache")
            _clean_caches(cache)


def grib_filter(change: Change, path: str) -> bool:
    """Make sure files we ingest are grib files."""
    if change == Change.deleted:
        # Can't process a deleted file
        return False
    if os.path.splitext(path)[1] in GRIB_ENDINGS:
        return True
    return False


def grib_file_watch(
    watch_dir: str, debounce: int | None = None, recursive=False
) -> Generator[str, None, None]:
    """Get grib files by watching a directory for changes.

    Args:
        watch_dir (str): The directory to watch for grib files.
        debounce (int | None): How long to wait to group file changes together before yielding.
        recursive (bool): Watch directories under watch_dir recursively?

    Returns:
        Generator[str, None, None]: A grib file path generator.
    """
    debounce = debounce or DEFAULT_WATCH_DEBOUNCE
    LOG.info(
        "Watching directory %s for gribs with debounce=%d, recursive=%s",
        watch_dir,
        debounce,
        recursive,
    )
    for changes in watch(
        watch_dir,
        watch_filter=grib_filter,
        debounce=debounce,
        recursive=recursive,
    ):
        for _, path in changes:
            yield path


def publish_message(msg: GribPayload) -> None:
    """Publish a message to RabbitMQ servers using the ssec_amqp package.
    This requires previous setup using ssec_amqp.api.connect().
    """
    route_key = GRIB_ROUTE_KEY_FMT.format(
        xcd_model=msg.get("xcd_model"), xcd_model_id=msg.get("xcd_model_id")
    )
    status = mq.publish(msg, route_key=route_key)
    LOG.info("Published to %s. status: %s", route_key, status)


def dump_message(msg: GribPayload) -> None:
    """Print a message to stdout."""
    print(json.dumps(msg), flush=True)
